
// (c) COPYRIGHT URI/MIT 1996, 1999
// Please read the full copyright statement in the file COPYRIGHT.
//
// Authors:
//      Dan Holloway <d.holloway@gso.uri.edu> dan

// Implementation for JGInt32. See JGByte.cc
//
// jhrg 1/12/95

#include "config_jg.h"

#ifdef __GNUG__
#pragma implementation
#endif

#include "Error.h"
#include "JGInt32.h"
#include "jgofs.h"

Int32 *
NewInt32(const string &n)
{
    return new JGInt32(n);
}

JGInt32::JGInt32(const string &n) : Int32(n)
{
    _missing_val = INT_MAX;
}

BaseType *
JGInt32::ptr_duplicate()
{
    return new JGInt32(*this);
}

bool
JGInt32::read(const string &dataset)
{
  for (int j = 0; j < nVars; ++j) {
    if (name() == varName[j]) {
      char *errptr;
      dods_int32 df = strtol(varValue[j], &errptr, 0);
      if ( errptr != (char *)&(varValue[j]) )
	val2buf(&df);
      else 
	val2buf(&_missing_val);	      
      set_read_p(true);
      return true;	// found so return indicating no more to do.
    }
  }

  // If we're here, the variable was not found. throw and error.
  string s = (string)"Variable " + name() 
    + " wasn't located in JGOFS object " + dataset + ".";
  throw Error(unknown_error, s);
}

dods_int32
JGInt32::get_missing_value() const
{
    return _missing_val;
}

void
JGInt32::set_missing_value(const dods_int32 &v)
{
    _missing_val = v;
}

// $Log: JGInt32.cc,v $
// Revision 1.10.4.2  2001/08/21 15:30:13  dan
// Updated names of the set/get accessor methods for modifying the missing_value
// member field, to make the naming convention consistent with the current
// DAP usage.
//
// Revision 1.10.4.1  2001/05/31 19:13:45  dan
// Modified return value from read method, now returns true.
//
// Revision 1.10  2000/10/10 23:36:14  jimg
// Added exception processing code.
// Added throws to the read methods.
// Changed the definition of read to match the dap library.
// Moved CVS Log entries to the end of files.
//
// Revision 1.9  1999/11/05 01:17:23  jimg
// Result of merge with release-3-1-0
//
// Revision 1.8.6.2  1999/10/30 19:43:35  dan
// Added explicit (char *) cast in error-checking within read().
//
// Revision 1.8.6.1  1999/10/24 18:29:44  dan
// Modified to support type conversion from JGOFS
// string variables to dods_int32.
//
// Revision 1.8  1999/05/05 04:38:00  brent
// convert String.h code to new stanard lib <string>
//
// Revision 1.7  1998/08/06 16:32:34  jimg
// Fixed misuse of the read(...) member function. Return true if more data
// is to be read, false is if not and error if an error is detected
//
// Revision 1.6  1996/11/12 16:55:19  jimg
// Added config_jg.h include
//
// Revision 1.5  1996/09/17 00:01:58  jimg
// Misc changes associated with merging branch code to the trunk.
//
// Revision 1.4.6.1.2.1  1996/06/26 22:17:56  jimg
// Dans changes for 2.05.
//
// Revision 1.4  1995/10/11  23:06:10  jimg
// These files are the result of merging Dan's changes into the main
// trunk. However, the archive was messed up in the process, so it may be hard
// to trace back to the original versions of some of these files.
//
// Revision 1.3  1995/07/09  21:39:50  jimg
// Added copyright notice.
//
// Revision 1.2  1995/03/14  01:23:27  jimg
// Modified so that these files are a proper subclass of the DAP classes
// Byte, ..., Grid.
// Added assert macros.
// Added Jgofs.h external declarations file.
//
// Revision 1.1  1995/02/16  19:01:16  dan
// JGOFS sub-class definitions for Int32.cc
//
// Revision 1.2  1995/01/19  21:58:57  jimg
// Added read_val from dummy_read.cc to the sample set of sub-class
// implementations.
// Changed the declaration of readVal in BaseType so that it names the
// mfunc read_val (to be consistant with the other mfunc names).
// Removed the unnecessary duplicate declaration of the abstract virtual
// mfuncs read and (now) read_val from the classes Byte, ... Grid. The
// declaration in BaseType is sufficient along with the decl and definition
// in the *.cc,h files which contain the subclasses for Byte, ..., Grid.
//
// Revision 1.1  1995/01/19  20:20:47  jimg
// Created as an example of subclassing the class hierarchy rooted at
// BaseType.
//
