// (c) COPYRIGHT URI/MIT 1996, 1999
// Please read the full copyright statement in the file COPYRIGHT.
//
// Authors:
//      Dan Holloway <d.holloway@gso.uri.edu> dan

// This file contains functions which read the variables and their attributes
// from a netcdf file and build the in-memeory DAS. These functions form the
// core of the server-side software necessary to extract the DAS from a
// netcdf data file.
//
// It also contains test code which will print the in-memeory DAS to
// stdout. It uses both the DAS class as well as the netcdf library.
// In addition, parts of these functions were taken from the netcdf program
// ncdump, from the netcdf standard distribution (ver 2.3.2)
//
// jhrg 9/23/94

#include "config_jg.h"

static char rcsid[] not_used ={"$Id: jgdas.cc,v 1.17.4.3 2002/06/21 00:31:40 jimg Exp $"};

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>

#include <iostream>

#include "cgi_util.h"
#include "util.h"
#include "DAS.h"
#include "dods-limits.h"
#include "jgofs.h"

// These are used as the return values for print_type().

static const char STRING[]="String";
static const char BYTE[]="Byte";
static const char INT16[]="Int16";
static const char UINT16[]="Uint16";
static const char INT32[]="Int32";
static const char UINT32[]="Uint32";
static const char FLOAT32[]="Float32";
static const char FLOAT64[]="Float64";

// Used by ErrMsgT

static char Msgt[255];
extern void ErrMsgT(const char *Msgt);

// Given a pointer to the jgofs variable id and a pointer to an attribute
// table pointer, read the attributes and store their names and values in the
// attribute table.
//
// NB: currently values are stored only as strings.
//
// Returns: Always returns true.

bool
read_attributes(int *v,  AttrTable *at, string &error)
{
    char attrname[TOKEN];

    AttrType var_attr, var_miss;
    string return_attr = "", dods_var = varName[*v];
    char attributeStr[TOKEN];
    
    unsigned char  missing_byte_value;
    short          missing_int16_value;
    unsigned short missing_uint16_value;
    long           missing_int32_value;
    unsigned long  missing_uint32_value; 
    float          missing_float32_value;
    double         missing_float64_value;
    string         missing_str_value;

    return_attr = at->get_attr("DODS_Type");
    downcase(return_attr);

    if (return_attr.find("container") != return_attr.npos)
      var_attr = Attr_container;
    else if (return_attr.find("byte") != return_attr.npos)
      var_attr = Attr_byte;
    else if (return_attr.find("uint16") != return_attr.npos)
      var_attr = Attr_uint16;
    else if (return_attr.find("int16") != return_attr.npos)
      var_attr = Attr_int16;
    else if (return_attr.find("uint32") != return_attr.npos)
      var_attr = Attr_uint32;
    else if (return_attr.find("int32") != return_attr.npos)
      var_attr = Attr_int32;
    else if (return_attr.find("float32") != return_attr.npos)
      var_attr = Attr_float32;
    else if (return_attr.find("float64") != return_attr.npos)
      var_attr = Attr_float64;
    else if (return_attr.find("string") != return_attr.npos)
      var_attr = Attr_string;
    else if (return_attr.find("url") != return_attr.npos)
      var_attr = Attr_url;
    else 
      var_attr = Attr_unknown;
    
    switch (var_attr) {
      
    case Attr_byte:
      var_miss = at->get_attr_type("missing_value");
      if ( var_miss == Attr_unknown ) {
	// No Missing_Value specification, use default value.
	missing_byte_value = DODS_UCHAR_MAX;
	sprintf(attributeStr, "%d", missing_byte_value);
	at->append_attr("missing_value", "Byte", attributeStr);
      }
      else if ( var_miss != Attr_byte ) {
	// Error condition, unless unspecified or 'byte' cannot set missing value
	// to one specified in the ancillary DAS.
	error += "\"JGOFS/DAS - Variable = " + dods_var + ", DODS_Type attr = " + return_attr +  + ", if provided in ancillary DAS file, missing_value attribute must be type BYTE" + "\"";
	return FALSE;
      }
      break;

    case Attr_int16:
      var_miss = at->get_attr_type("missing_value");
      if ( var_miss == Attr_unknown ) {
	// No Missing_Value specification, use default value.
	missing_int16_value = DODS_SHRT_MAX;
	sprintf(attributeStr, "%d", missing_int16_value);
	at->append_attr("missing_value", "Int16", attributeStr);
      }
      else if ( var_miss != Attr_int16 ) {
	// Error condition, unless unspecified or 'int16' cannot set missing value
	// to one specified in the ancillary DAS.
	error += "\"JGOFS/DAS - Variable = " + dods_var + ", DODS_Type attr = " + return_attr +  + ", if provided in ancillary DAS file, missing_value attribute must be type INT16" + "\"";
	return FALSE;
      }
      break;

    case Attr_uint16:
      var_miss = at->get_attr_type("missing_value");
      if ( var_miss == Attr_unknown ) {
	// No Missing_Value specification, use default value.
	missing_uint16_value = DODS_USHRT_MAX;
	sprintf(attributeStr, "%u", missing_uint16_value);
	at->append_attr("missing_value", "UInt16", attributeStr);
      }
      else if ( var_miss != Attr_uint16 ) {
	// Error condition, unless unspecified or 'uint16' cannot set missing value
	// to one specified in the ancillary DAS.
	error += "\"JGOFS/DAS - Variable = " + dods_var + ", DODS_Type attr = " + return_attr +  + ", if provided in ancillary DAS file, missing_value attribute must be type UINT16" + "\"";
	return FALSE;
      }
      break;

    case Attr_int32: 
      var_miss = at->get_attr_type("missing_value");
      if ( var_miss == Attr_unknown ) {
	// No Missing_Value specification, use default value.
	missing_int32_value = DODS_INT_MAX;
	sprintf(attributeStr, "%d", missing_int32_value);
	at->append_attr("missing_value", "Int32", attributeStr);
      }
      else if ( var_miss != Attr_int32 ) {
	// Error condition, unless unspecified or 'int32' cannot set missing value
	// to one specified in the ancillary DAS.
	error += "\"JGOFS/DAS - Variable = " + dods_var + ", DODS_Type attr = " + return_attr +  + ", if provided in ancillary DAS file, missing_value attribute must be type INT32" + "\"";
	return FALSE;
      }
      break;

    case Attr_uint32: 
      var_miss = at->get_attr_type("missing_value");
      if ( var_miss == Attr_unknown ) {
	// No Missing_Value specification, use default value.
	missing_uint32_value = DODS_UINT_MAX;
	int stat = sprintf(attributeStr, "%u", missing_uint32_value);
	at->append_attr("missing_value", "UInt32", attributeStr);
      }
      else if ( var_miss != Attr_uint32 ) {
	// Error condition, unless unspecified or 'uint32' cannot set missing value
	// to one specified in the ancillary DAS.
	error += "\"JGOFS/DAS - Variable = " + dods_var + ", DODS_Type attr = " + return_attr +  + ", if provided in ancillary DAS file, missing_value attribute must be type UINT32" + "\"";
	return FALSE;
      }
      break;

    case Attr_float32:
      var_miss = at->get_attr_type("missing_value");
      if ( var_miss == Attr_unknown ) {
	// No Missing_Value specification, use default value.
	missing_float32_value = DODS_FLT_MAX;
	sprintf(attributeStr, "%e", missing_float32_value);
	at->append_attr("missing_value", "Float32", attributeStr);
      }
      else if ( var_miss != Attr_float32 ) {
	// Error condition, unless unspecified or 'float32' cannot set missing value
	// to one specified in the ancillary DAS.
	error += "\"JGOFS/DAS - Variable = " + dods_var + ", DODS_Type attr = " + return_attr +  + ", if provided in ancillary DAS file, missing_value attribute must be type FLOAT32" + "\"";
	return FALSE;
      }
      break;
      
    case Attr_float64: 
      var_miss = at->get_attr_type("missing_value");
      if ( var_miss == Attr_unknown ) {
	// No Missing_Value specification, use default value.
	missing_float64_value = DODS_DBL_MAX;
	sprintf(attributeStr, "%e", missing_float64_value);
	at->append_attr("missing_value", "Float64", attributeStr);
      }
      else if ( var_miss != Attr_float64 ) {
	// Error condition, unless unspecified or 'float64' cannot set missing value
	// to one specified in the ancillary DAS.
	error += "\"JGOFS/DAS - Variable = " + dods_var + ", DODS_Type attr = " + return_attr +  + ", if provided in ancillary DAS file, missing_value attribute must be type FLOAT64" + "\"";
	return FALSE;
      }
      break;
     
    case Attr_string: 
      var_miss = at->get_attr_type("missing_value");
      if ( var_miss == Attr_unknown ) {
	// No Missing_Value specification, use default value.
#if 0
	missing_str_value = "nd";
	sprintf(attributeStr, "\"nd\"", missing_str_value);
#endif
	strcpy(attributeStr, "\"nd\"");
	at->append_attr("missing_value", "String", attributeStr);
      }
      else if ( var_miss != Attr_string ) {
	// Error condition, unless unspecified or 'String' cannot set missing value
	// to one specified in the ancillary DAS.
	error += "\"JGOFS/DAS - Variable = " + dods_var + ", DODS_Type attr = " + return_attr +  + ", if provided in ancillary DAS file, missing_value attribute must be type STRING" + "\"";
	return FALSE;
      }
      break;

    case Attr_url: 
    case Attr_container:
    case Attr_unknown: 
      break;

   }

    int idx;
    string attr = "";
    string name = "";
    string val = "";
    string quotedVal = "";

    while( ljdbattributes_(&unit, v, attrname) ) {

      attr = attrname;
      if (((idx = attr.find("=")) != string::npos) || ((idx = attr.find("-")) != string::npos)) {
	name = attr.substr(0,idx);
	val = attr.substr(idx+1, attr.length()-idx+1);

	quotedVal = "\"" + val + "\"";
	at->append_attr(name, "String", quotedVal);
      }
      else {
	quotedVal = "\"" + attr + "\"";
	at->append_attr("units", "STRING", quotedVal);
      }
    }
    
    return true;
}

// Given a reference to an instance of class DAS and a filename that referrs
// to a JGOFS object, read the JGOFS object and extract all the attributes
// of each of its variables. Add the variables and their attributes to the
// instance of DAS.
//
// Returns: false if an error accessing the JGOFS object was detected, true
// otherwise. 

bool
create_das(DAS &das, string &obj_path, string &met_path, string &jg_name, string &error)
{
    bool status;
    char msg[1024];
    char jg_name_c[256], obj_path_c[256], met_path_c[256];
    size_t pos;

    strncpy(jg_name_c, jg_name.c_str(), 255);
    jg_name_c[255] = '\0';	// for the truly paranoid...
    strncpy(obj_path_c, obj_path.c_str(), 255);
    obj_path_c[255] = '\0';	
    strncpy(met_path_c, met_path.c_str(), 255);
    met_path_c[255] = '\0';	


    // Replaced -99 with -NVALS. jhrg 11/8/96
    nVars = -(NVALS);
    int nLvls = ljdbopen_(&unit, obj_path_c, met_path_c, jg_name_c, varName, &varSize, &nVars);
    AttrTable *attr_table_ptr;

    if (nLvls < 0) {
	char msg[1024];

        if (ljdberror_(&unit, &msg)) 
	    error += "\"Could not open " + jg_name + ": " + msg + "\"";
	else 
	    error += "\"JGOFS/DDS - could not open object: " +  jg_name + "\"";

	return false;
    }

    // `v' is the variable id passed to read_attributes().

    for (int v = 0; v < nVars; ++v) {
	attr_table_ptr = das.get_table(varName[v]);
	if (!attr_table_ptr)
	    attr_table_ptr = das.add_table(varName[v], new AttrTable);
      
	if (!read_attributes(&v, attr_table_ptr, error))
	    return false;
    }

    return true;
}

// $Log: jgdas.cc,v $
// Revision 1.17.4.3  2002/06/21 00:31:40  jimg
// I changed many files throughout the source so that the 'make World' build
// works with the new versions of Connect and libdap++ that use libcurl.
// Most of these changes are either to Makefiles, configure scripts or to
// the headers included by various C++ files. In a few places the oddities
// of libwww forced us to hack up code and I've undone those and some of the
// clients had code that supported libwww's generous tracing capabilities
// (that's one part of libwww I'll miss); I had to remove support for that.
// Once this code compiles and more work is done on Connect, I'll return to
// each of these changes and polish them.
//
// Revision 1.17.4.2  2002/01/15 21:05:08  jimg
// I switched from the numeric constants in values.h/limits.h to those in
// dods-limits.h.
//
// Revision 1.17.4.1  2001/08/21 15:35:01  dan
// Extensive modifications to handle exceptions properly, and to clean up
// the object dictionary search logic. Added relative path searching to
// locate the object dictionaries, as per request U.S. JGOFS Office.
//
// Modified read_attributes function to handle case properly when there
// is an ancillary DAS as well as inline JGOFS attributes.
//
// Revision 1.17  2000/10/10 23:36:14  jimg
// Added exception processing code.
// Added throws to the read methods.
// Changed the definition of read to match the dap library.
// Moved CVS Log entries to the end of files.
//
// Revision 1.16  1999/11/05 01:17:26  jimg
// Result of merge with release-3-1-0
//
// Revision 1.15.2.3  1999/11/03 01:30:44  jimg
// Fixed a warning for nd with Strings
//
// Revision 1.15.2.2  1999/10/30 19:46:10  dan
// Fixed bad return type when returning on error condition in new
// attribute handling code.
//
// Revision 1.15.2.1  1999/10/24 18:26:41  dan
// Modified to support data provider specifed return
// types other than String.  This new methodology uses
// the DODS_Type and missing_value attributes, which
// may be specified per variable in an ancillary DAS.
//
// Revision 1.15  1999/07/22 22:38:10  jimg
// Merged changes from the release-3-0-2 branch
//
// Revision 1.14.2.1  1999/06/07 17:30:28  dan
// Changed string class usage 'data()' to 'c_str()'.
//
// Revision 1.14  1999/05/25 16:31:51  dan
// Replaced char* with string in argument list, consistent with jgdds.cc
//
// Revision 1.13  1999/05/05 22:57:20  jimg
// Merged Brent's String --> string changes with Dan's changes
//
// Revision 1.12  1999/05/05 04:38:00  brent
// convert String.h code to new stanard lib <string>
//
// Revision 1.11  1998/02/05 20:14:38  jimg
// DODS now compiles with gcc 2.8.x
//
// Revision 1.10  1997/03/10 22:58:24  dan
// Added Error-handling.  Use new API routines added to local
// jgofs library to retrieve errors generated by the local methods.
//
// Revision 1.9  1996/12/18 17:44:38  jimg
// Removed dead code.
// Fixed the prototype of ErrMsgT() and uncommented its used in a conditional
// that tested for an error. All part of the long, continuing, saga of the
// PATH_INFO bug.
//
// Revision 1.8  1996/11/09 00:41:59  jimg
// Increased the number of variable that can be processed from 99 to NVALS
// (250).
//
// Revision 1.7  1996/09/17 00:04:12  jimg
// Misc changes associated with merging branch code to the trunk.
//
// Revision 1.6.2.1.2.1  1996/06/26 22:18:23  jimg
// Dans changes for 2.05.
//
// Revision 1.5  1995/10/11  23:56:10  jimg
// Added external symbols defined/declared in jgofs.h to lgofs/ljdb.c. Fixed
// the C constant TOKEN so that it is `40' everywhere.
//
// Revision 1.4  1995/10/11  23:06:28  jimg
// These files are the result of merging Dan's changes into the main
// trunk. However, the archive was messed up in the process, so it may be hard
// to trace back to the original versions of some of these files.
//
// Revision 1.3.2.1  1995/09/18  22:48:06  jimg
// Merged Dan's changes with the trunk version. This will be commited to the
// release-1-0-patches branch and then merged back into the trunk later.
//
// Revision 1.3  1995/07/09  21:40:04  jimg
// Added copyright notice.
//
// Revision 1.2  1995/03/14  02:17:25  jimg
// Modified so that the new netio library is used.
// Modified so that C++ Strings are used in place of C strings.
// Removed old code copied from the dods netcdf sources.
//
// Revision 1.1  1995/01/09  16:04:52  dan
// These files constitute the JGOFS server and client library code
// for use with the DODS api.
//
// Revision 1.1  1994/09/27  23:19:37  jimg
// First version of the netcdf software which used libdas++.a.
//
