
/*
   This copy of dct.c (renamed ldct.c) contains exactly the same source code
   as the dct.c file dustributed with JGOFS (unknow version of circa June
   1995) *except* that each external symbol has been renamed to
   `l<symbol_name>'. Thus the function `scandct' in the original source file is
   now `lscandct'.

   jhrg 9/18/95
 */

/* Well, not anymore. The local version of dct.c (renamed ldct.c) has been
   modified to support dereferencing of multiple object dictionaries, using
   a relative-path to the JGOFS_OBJECT location, prepended to the filename
   part of the URL.  For example, the filename 'ArabianSea/test1' would
   parse to relative-path = 'ArabianSea', object = 'test1'.   The relative-path
   would be appended to the standard JGOFS_OBJECT location in the dictionary
   search logic, thus allowing multiple object dictionaries to be accessible 
   from a single dods-jgofs server.  (As per request U.S.JGOFS Office)

   danh 7/31/01
 */

/*
 * $Log: ldct.c,v $
 * Revision 1.5.14.1  2001/08/21 15:39:25  dan
 * Modified dctsearch to support relative pathnames for locating local
 * object dictionaries.   Migrated DODS-specific search logic to a new
 * module ../jg_path.cc to keep all DODS-specific search logic in a single
 * place.
 *
 * Revision 1.5  1999/03/17 22:05:29  dan
 * Modified HOME_DIR_HACK, added filesystem path to JGOFS methods,
 * previously assumed relative to 'jgofs' home-dir.
 *
 * Revision 1.4  1999/03/02 19:23:59  dan
 * Modified HOME_DIR_HACK code section to return error messages
 * to the calling routines, and moved code section to the beginning
 * of dctsearch.
 *
 * Added the ability to check for shell environment variables
 * JGOFS_OBJECT and JGOFS_METHOD which can be used to circumvent
 * the /etc/passwd requirement for a 'jgofs' user.  If these
 * don't exist, and there isn't a user 'jgofs' defined an error
 * will be reported to the server, and reported in the httpd logs.
 *
 * Revision 1.3  1996/11/09 00:45:08  jimg
 * Increased MAX_BUF from 1025 to 2048.
 *
 * Revision 1.2  1995/10/17 21:44:17  jimg
 * Added the Home Directory Hack; this uses the jgofs user's home
 * directory instead of a compiled-in pathname to find data objects.
 * This way, DODS can distribute binaries and *not* require building the
 * software on each machine (and require a specific dir or env variable
 * be used for the paths (but the username must be defined...)).
 * */

/*
   #define OBJECT_ROOT "/usr/local/jgofs/objects"
   #define METHOD_ROOT "/usr/local/jgofs/methods/"
*/

/*
   #define DEBUG
*/
#include <stdio.h>
#include <string.h>
char *getenv();
#include <unistd.h>
#include <pwd.h>

/* Changed MAX_BUF from 1025 to 2048 to accomodate NSCAT data. jhrg 11/8/96 */
#define MAX_BUF 2048

/* 
   The HOME_DIR_HACK uses the current machine's /etc/passwd file (or whatever
   getpwnam(3) searches) to find the JGOFS home directory. We assume that the
   jgofs object root directory is a subdir of the home dir called
   `objects'. This seems to be OK to do since the Makefile uses and
   equivalent assumption when it builds the JGOFS API. Similar code is used
   to find the methods directory.

   I added this code because the previous method set OBJECT_ROOT and
   METHOD_ROOT at compile time. Since DODS will distribute binaries
   (i.e., precompiled libraries, servers and some demo clients) the paths
   compiled in on the machines used to build those binaries will be the paths
   searched. This hack figures out likely values for these paths at run time,
   so the values no longer depend on the machine used to compile the
   software. 

   10/13/95 jhrg 
 */

#define HOME_DIR_HACK 1
#define FALSE 0
#define TRUE 1

static char rcsid[] = {"$Id: ldct.c,v 1.5.14.1 2001/08/21 15:39:25 dan Exp $"};

int lscandct(s, t)
char *s, *t;
{
    FILE *fdic;
    char tdic[MAX_BUF], dctname[MAX_BUF], *sp;
    int nok;
#ifdef DEBUG
    printf("Searching >%s< for >%s<\n", s, t);
#endif
    sp = strrchr(t, '/');
    if (sp) {
	*(sp++) = 0;
	strcpy(dctname, t);
	strcat(dctname, "/");
	strcat(dctname, s);
    } else {
	strcpy(dctname, s);
	sp = t;
    }
    strcat(sp, "=");
#ifdef DEBUG
    printf("Open >%s< and look for >%s<\n", dctname, sp);
#endif
    fdic = fopen(dctname, "r");
    if (fdic == NULL)
	return 0;

    nok = 1;
    while (!feof(fdic) && nok) {
	fgets(tdic, MAX_BUF - 1, fdic);
	if (strstr(tdic, sp) == tdic)
	    nok = 0;
    }
    fclose(fdic);
    if (nok)
	return 0;

    strcpy(t, strchr(tdic, '=') + 1);
    t[strlen(t) - 1] = 0;
    return 1;
}

lmergeparams(req, dct)
char *req, *dct;
{
    char tmp[MAX_BUF], *j, *k;
    int i, dctopt, reqproj, nb;

    /* no params in dictionary */
    if (dct == NULL)
	return;

    /* no params requested */
    if (*req == 0) {
	j = dct + 1;
	k = req;
	while (*j) {
	    if (*j == '[' || *j == ']')
		j++;
	    else
		*(k++) = *(j++);
	}
	*dct = 0;
	*(k - 1) = 0;
	return;
    }
    /* dictionary params first */

    nb = 0;
    dctopt = 1;
    j = dct;
    k = tmp;
    while (*(++j)) {
	switch (*j) {
	case '[':
	    nb++;
	    dctopt = 0;
	    break;
	case ']':
	    nb--;
	    break;
	default:
	    if (nb == 0)
		*(k++) = *j;
	    break;
	}
    }
    if (k > tmp)
	if (*(k - 1) == ')')
	    *(k - 1) = ',';

    j = req - 1;
    reqproj = 0;
    nb = 0;
    i = 1;
    while (*(++j)) {
	*(k++) = *j;
	switch (*j) {
	case '(':
	    nb++;
	    break;
	case ')':
	    nb--;
	    break;
	case '<':
	case '>':
	case '=':
	    i = 0;
	    break;
	case ',':
	    if (nb == 0 && i == 1)
		reqproj = 1;
	    i = 1;
	    break;
	}
    }
    *k = 0;

    if (reqproj || dctopt) {
	while (j = strstr(tmp, ",,"))
	    strcpy(j, j + 1);
	strcpy(req, tmp);
	*dct = 0;
	return;
    }
    j = dct;
    nb = 0;
    while (*(++j)) {
	switch (*j) {
	case '[':
	    nb++;
	    break;
	case ']':
	    nb--;
	    break;
	default:
	    if (nb)
		*(k++) = *j;
	    break;
	}
    }
    if (k > tmp) {
	if (*(k - 1) == ',')
	    *(k - 1) = 0;
	else
	    *k = 0;
	while (j = strstr(tmp, ",,"))
	    strcpy(j, j + 1);
	strcpy(req, tmp);
    } else
	*req = 0;
    *dct = 0;
    return;
}

#include <sys/stat.h>

int test_jg_path(pathname)
char *pathname;
{
    struct stat buf;
    int i;

    if ( lstat(pathname,&buf) < 0 ) return 0;
    else {
      if ( S_ISDIR(buf.st_mode) || S_ISLNK(buf.st_mode)) return 1;
      return 0;
    }
}

int ldctsearch(object_path, method_path, obj, method, params, localflag)
char *object_path, *method_path, *obj, *method, *params;
int localflag;
{
    char pathreq[MAX_BUF], metreq[MAX_BUF], parreq[MAX_BUF];
    char tmp[MAX_BUF], *j;
    char *environment_jg_object, *environment_jg_method;
    char OBJECT_PATH[256], METHOD_PATH[256];

    int object_error, method_error;
    int dctcount;
    int dctnfound;
    struct passwd *pwd_entry;	/* Added 10/13/95 jhrg */

    strcpy(pathreq, obj);
    strcpy(parreq, params);

    *method = 0;
    *params = 0;
    
    dctcount = 8;

    while (dctcount) {

	dctnfound = 1;

	/* //machine ... */
#ifdef DEBUG
	printf("test // ... %s**%s\n", pathreq, parreq);
#endif
	if (strstr(pathreq, "//") == pathreq) {
	    strcpy(method, pathreq);
	    strcpy(params, parreq);
	    return 1;
	}
	/* is path executable? */
	if (localflag) {
#ifdef DEBUG
	    printf("test local meth... %s**%s\n", pathreq, parreq);
#endif
	    if (access(pathreq, X_OK) == 0) {
	        strcpy(method, method_path);
		strcat(method, pathreq);
		strcpy(params, parreq);
		return 1;
	    }
	}
	/* check for object dictionary */

	if (localflag) {
	    strcpy(tmp, pathreq);
#ifdef DEBUG
	    printf("test local obj... %s**%s\n", tmp, parreq);
#endif
	    if (lscandct(".objects", tmp)) {
		j = strchr(tmp, '(');
		if (j)
		    lmergeparams(parreq, j);
		if (strstr(tmp, "//") == tmp) {
		    strcpy(method, tmp);
		    strcpy(params, parreq);
		    return 1;
		}
		strcpy(pathreq, tmp);
		dctnfound = 0;
	    }
	}
	/* check system local objects dictionary */

	if (dctnfound) {
	    if (*pathreq == '/') {
		strcpy(tmp, object_path);
	    }
	    else {
	      if (*pathreq) {
		strcpy(tmp, object_path);
		strcat(tmp, "/");
		strcat(tmp, pathreq);
	      }
	    }
#ifdef DEBUG
	    printf("test sys local ... %s**%s\n", tmp, parreq);
#endif
	    if (lscandct(".objects", tmp)) {
	      j = strchr(tmp, '(');
	      if (j)
		lmergeparams(parreq, j);
	      if (strstr(tmp, "//") == tmp) {
		strcpy(method, tmp);
		strcpy(params, parreq);
		return 1;
	      }
	      strcpy(pathreq, tmp);
	      dctnfound = 0;
	    }
	}

	/* check system dictionary */

	if (dctnfound) {
	    if (*pathreq == '/') {
	      strcpy(tmp, object_path);
	    }
	    else {
	      if (*pathreq) {
		strcpy(tmp, object_path);
		strcat(tmp, "/");
		strcat(tmp, pathreq);
	      }
	    }
#ifdef DEBUG
	    printf("test sys global ... %s**%s\n", tmp, parreq);
#endif
	    if (lscandct(".remoteobjects", tmp)) {
	      j = strchr(tmp, '(');
	      if (j)
		lmergeparams(parreq, j);
	      if (strstr(method, "//") == method) {
		strcpy(method, tmp);
		strcpy(params, parreq);
		return 1;
	      }
	      strcpy(pathreq, tmp);
	      dctnfound = 0;
	    }
	}

	/* check for system method */

	if (dctnfound) {
	  if ( localflag || dctcount < 8 ) {
	    strcpy(tmp, method_path);
	    strcat(tmp, pathreq);
#ifdef DEBUG
	  printf("test sys meth ... %s**%s\n", tmp, parreq);
#endif
	    if (access(tmp, X_OK) == 0) {
	      strcpy(method, tmp);
	      strcpy(params, parreq);
	      return 1;
	    }
	  }
	}

        /* local file - use default method */

	if(dctnfound && localflag) {

#ifdef DEBUG
    printf("test def file ... %s**\n",pathreq);
#endif
          if(access(pathreq,R_OK)==0) {
	    strcpy(tmp,"(");
	    strcat(tmp,pathreq);
	    strcat(tmp,")");
	    lmergeparams(parreq,tmp);
	    strcpy(pathreq,"def");
	    dctnfound=0;
	  }
	}

	/* Error */
	if (dctnfound) {
	    strcpy(method, "Cannot be found in dictionary:");
	    return 0;
	}
	dctcount--;
    }
    strcpy(method, "Too many iterations in dictionary - check for loops:");
    return 0;
}

/*
   main(argc,argv)
   int argc;
   char *argv[];
   {
   char met[1024],par[1024];
   int i;
   if(ldctsearch(argv[1],met,par,1))
   printf("ldctsearch = %s**%s\n",met,par);
   else printf("Not found\n");
   }
*/










